"""Define the factories of the program."""

from typing import Any

from pydantic_factories import ModelFactory

from pass_collaborate.model import auth


class GroupFactory(ModelFactory[Any]):
    """Define factory for the Group model."""

    __model__ = auth.Group


class UserFactory(ModelFactory[Any]):
    """Define factory for the Group model."""

    __model__ = auth.User



"""Define program exceptions."""


class NotFoundError(Exception):
    """Model the exception of not finding something."""


class TooManyError(Exception):
    """Model the exception of finding too much."""


class DecryptionError(Exception):
    """Model the exception of problems when decrypting a file."""


class EncryptionError(Exception):
    """Model the exception of problems when encrypting a file."""



"""Test the implementation of the group command line command."""

import logging
import re
from pathlib import Path
from typing import TYPE_CHECKING, List

import pytest
from _pytest.logging import LogCaptureFixture
from typer.testing import CliRunner

from pass_collaborate.entrypoints.cli import app
from pass_collaborate.model.auth import Group, User

from ..factories import GroupFactory

if TYPE_CHECKING:
    from pass_collaborate.model.auth import AuthStore
    from pass_collaborate.model.pass_ import PassStore


def test_group_add(
    cli_runner: CliRunner,
    pass_: "PassStore",
    admin: "User",
    developer: "User",
    attacker: "User",
) -> None:
    """
    Given: A configured environment
    When: calling group add command with an email, a user name and a key id
    Then: A group is added and the data stored in the auth store are the emails
    """
    pass_.auth.add_user(developer.name, developer.key, developer.email)
    pass_.auth.add_user(attacker.name, attacker.key, attacker.email)

    result = cli_runner.invoke(
        app, ["group", "add", "test_group", developer.email, admin.name, attacker.key]
    )

    assert result.exit_code == 0
    pass_.auth.reload()
    saved_group, _ = pass_.auth.get_group("test_group")
    assert saved_group == Group(
        name="test_group", users=[developer.email, admin.email, attacker.email]
    )


def test_group_list(cli_runner: CliRunner, pass_: "PassStore") -> None:
    """
    Given: A configured environment and a group added
    When: calling group list command
    Then: The groups are listed
    """
    pass_.auth.add_group("test_group")

    result = cli_runner.invoke(app, ["group", "list"])

    assert result.exit_code == 0
    assert "test_group" in result.stdout


@pytest.mark.parametrize(
    "group",
    [
        GroupFactory.build(users=[]),
        GroupFactory.build(users=["developer@example.org"]),
        GroupFactory.build(users=["developer@example.org", "admin@example.org"]),
    ],
)
def test_group_show(
    cli_runner: CliRunner,
    pass_: "PassStore",
    developer: "User",
    admin: "User",
    group: Group,
) -> None:
    """
    Given: A configured environment and a group added
    When: calling group show command
    Then: The group information is shown
    """
    pass_.auth.add_user(developer.name, developer.key, developer.email)
    pass_.auth.add_group(name=group.name, user_ids=group.users)  # type: ignore

    result = cli_runner.invoke(app, ["group", "show", group.name])

    assert result.exit_code == 0
    assert re.search(rf"{group.name}", result.stdout)
    assert group.users is not None
    if len(group.users) > 0:
        assert re.search(
            rf".*{developer.name}: {developer.email} \({developer.key}\).*",
            result.stdout,
        )
    if len(group.users) > 1:
        assert re.search(
            rf".*{admin.name}: {admin.email} \({admin.key}\).*", result.stdout
        )


@pytest.mark.parametrize(
    "entity",
    [
        "developers",
        "developer@example.org",
        "developer",
        "8DFE8782CD025ED6220D305115575911602DDD94",
    ],
)
def test_group_authorize_a_directory(
    cli_runner: CliRunner,
    pass_: "PassStore",
    pass_dev: "PassStore",
    entity: str,
    developer: User,
) -> None:
    """
    Given: A configured environment and a group added. With no `.gpg-id` file
        in the directory we want to change the permissions.
    When: calling group authorize command on a subdirectory either with a group
        name or a user identifier.
    Then: The group members are authorized to access the data and the .gpg-id
        file of the directory contains the new key.
    """
    gpg_id = Path(pass_.store_dir / "web" / ".gpg-id")
    assert not gpg_id.is_file()
    pass_.auth.add_user(name=developer.name, email=developer.email, key=developer.key)
    pass_.auth.add_group(name="developers", user_ids=["developer@example.org"])
    # Check that the permissions are right
    for element in ["web", "database", "bastion"]:
        assert pass_.has_access(element)
        assert not pass_.has_access(element, identifier=developer.email)
    for environment in ("production", "staging"):
        assert pass_.can_decrypt(pass_.path(f"web/{environment}"))
        assert not pass_dev.can_decrypt(pass_.path(f"web/{environment}"))

    result = cli_runner.invoke(app, ["group", "authorize", entity, "web"])

    assert result.exit_code == 0
    pass_.auth.reload()
    assert pass_.has_access("web", identifier=developer.email)
    for environment in ("production", "staging"):
        assert pass_.can_decrypt(pass_.path(f"web/{environment}"))
        assert pass_dev.can_decrypt(pass_.path(f"web/{environment}"))
    # Check that the permissions of the rest of the store have not changed.
    assert pass_.has_access("web")
    for element in ["database", "bastion"]:
        assert pass_.has_access(element)
        assert not pass_.has_access(element, identifier=developer.email)
    assert "8DFE8782CD025ED6220D305115575911602DDD94" in gpg_id.read_text(
        encoding="utf-8"
    )


def test_group_authorize_cant_authorize_file(cli_runner: CliRunner) -> None:
    """
    Given: A configured environment
    When: Trying to authorize a file
    Then: An error is raised as we don't yet support giving granular
        permissions to files.
    """
    result = cli_runner.invoke(app, ["group", "authorize", "user", "bastion"])

    assert result.exit_code == 2
    assert "Changing access to a file is not yet supported" in result.stderr


def test_group_authorize_is_idempotent(
    cli_runner: CliRunner, pass_: "PassStore", admin: "User"
) -> None:
    """
    Given: A configured environment
    When: Trying to authorize a directory that is already authorized
    Then: It doesn't do anything as it's already authorized.
    """
    pass_.auth.add_group(name="admins", user_ids=[admin.email])
    pass_.change_access("web", ["admins"])
    assert pass_.auth.access["web/.gpg-id"] == ["admins"]

    result = cli_runner.invoke(app, ["group", "authorize", "admins", "web"])

    assert result.exit_code == 0
    pass_.auth.reload()
    assert pass_.auth.access["web/.gpg-id"] == ["admins"]


def test_group_authorize_can_ignore_parent(
    cli_runner: CliRunner, pass_: "PassStore", admin: "User", developer: "User"
) -> None:
    """
    Given: A configured environment
    When: Authorize a directory with the --ignore_parent flag
    Then: It ignores the permissions of the parent .gpg-id file
        and sets only the new one
    """
    pass_.auth.add_user(name=developer.name, email=developer.email, key=developer.key)
    pass_.auth.add_group(name="developers", user_ids=[developer.email])
    assert pass_.has_access("web", admin.name)
    assert not pass_.has_access("web", "developers")

    result = cli_runner.invoke(
        app, ["group", "authorize", "--ignore-parent", "developers", "web"]
    )

    assert result.exit_code == 0
    pass_.auth.reload()
    assert not pass_.has_access("web", admin.name)
    assert pass_.has_access("web", "developers")


@pytest.mark.parametrize(
    "arguments",
    [
        ["group", "add-users", "developer@example.org", "test_group"],
        [
            "group",
            "add-users",
            "developer@example.org",
            "admin@example.org",
            "test_group",
        ],
    ],
)
def test_group_add_users(
    cli_runner: CliRunner,
    pass_: "PassStore",
    arguments: List[str],
    developer: User,
    admin: User,
) -> None:
    """
    Given: A configured environment and an empty group
    When: adding users to a group
    Then: users are added
    """
    pass_.auth.add_user(name=developer.name, email=developer.email, key=developer.key)
    pass_.auth.add_group("test_group")

    result = cli_runner.invoke(app, arguments)

    assert result.exit_code == 0
    pass_.auth.reload()
    saved_group, saved_users = pass_.auth.get_group("test_group")
    assert saved_group.users is not None
    if "developer@example.org" in arguments:
        assert developer in saved_users
    if "admin@example.org" in arguments:
        assert admin in saved_users


def test_group_with_associated_passwords_add_users(
    cli_runner: CliRunner,
    pass_: "PassStore",
    pass_dev: "PassStore",
    developer: User,
    admin: User,
) -> None:
    """
    Given: A configured environment and a group authorized to some passwords
    When: adding users to a group
    Then: the new users are able to read the group passwords
    """
    pass_.auth.add_user(name=developer.name, email=developer.email, key=developer.key)
    pass_.auth.add_group(name="developers", user_ids=[admin.email])
    pass_.change_access(add_identifiers=["developers"], pass_dir_path="web")
    for environment in ("production", "staging"):
        assert pass_.can_decrypt(pass_.path(f"web/{environment}"))
        assert not pass_dev.can_decrypt(pass_.path(f"web/{environment}"))

    result = cli_runner.invoke(
        app, ["group", "add-users", developer.email, "developers"]
    )

    assert result.exit_code == 0
    for environment in ("production", "staging"):
        assert pass_.can_decrypt(pass_.path(f"web/{environment}"))
        assert pass_dev.can_decrypt(pass_.path(f"web/{environment}"))


def test_group_with_associated_passwords_in_gpg_id_file_add_users(
    cli_runner: CliRunner,
    pass_: "PassStore",
    pass_dev: "PassStore",
    developer: User,
    admin: User,
    attacker: User,
) -> None:
    """
    Given: A configured password store with a gpg_id file with the keys of
        developer, attacker and admin and a group `all` only with `admin`.
    When: adding users to a group
    Then: the new users are able to read the group passwords
    """
    # Create the .gpg-id file to have both keys and an empty auth store
    pass_.auth.add_user(name=developer.name, email=developer.email, key=developer.key)
    pass_.auth.add_user(name=attacker.name, email=attacker.email, key=attacker.key)
    pass_.auth.add_group(
        name="all", user_ids=[admin.email, developer.email, attacker.email]
    )
    pass_.change_access(add_identifiers=["all"], pass_dir_path=".")
    pass_.auth.groups = []
    pass_.auth.access = {}
    pass_.auth.save()
    pass_.auth.reload()
    # As the `all` group doesn't exist, both users are in the .gpg-id file
    # When we automatically load the information from the .gpg-id files it adds the
    # users keys individually
    assert admin.key in pass_.auth.access[".gpg-id"]
    assert developer.key in pass_.auth.access[".gpg-id"]
    assert attacker.key in pass_.auth.access[".gpg-id"]
    #
    # Simulate that we create the group
    pass_.auth.add_group(name="all", user_ids=[admin.email])
    pass_.change_access(add_identifiers=["all"], pass_dir_path=".")
    # When we create the group, the reference to admin.name is replaced by the group
    # it belongs to, all. But the developer.name it's still there as it's not part
    # of the group but the .gpg-id grants it access
    assert "all" in pass_.auth.access[".gpg-id"]
    assert developer.key in pass_.auth.access[".gpg-id"]
    assert attacker.key in pass_.auth.access[".gpg-id"]

    result = cli_runner.invoke(
        app, ["group", "add-users", developer.email, attacker.email, "all"]
    )

    assert result.exit_code == 0
    pass_.auth.reload()
    # Now that developer is part of `all` there should not be any reference to that
    # user in the access information.
    assert pass_.auth.access[".gpg-id"] == ["all"]


def test_group_remove_user_from_group(
    cli_runner: CliRunner,
    auth: "AuthStore",
    pass_: "PassStore",
    pass_dev: "PassStore",
    developer: User,
    admin: User,
) -> None:
    """
    Given: A configured environment and a group authorized to some passwords
    When: removing a user from the group
    Then: the removed user are not able to read the group passwords
    """
    pass_.auth.add_user(name=developer.name, email=developer.email, key=developer.key)
    pass_.auth.add_group(name="developers", user_ids=[admin.email, developer.email])
    pass_.change_access(add_identifiers=["developers"], pass_dir_path="web")
    for environment in ("production", "staging"):
        assert pass_.can_decrypt(pass_.path(f"web/{environment}"))
        assert pass_dev.can_decrypt(pass_.path(f"web/{environment}"))

    result = cli_runner.invoke(
        app, ["group", "remove-users", developer.email, "developers"]
    )

    assert result.exit_code == 0
    for environment in ("production", "staging"):
        assert pass_.can_decrypt(pass_.path(f"web/{environment}"))
        assert not pass_dev.can_decrypt(pass_.path(f"web/{environment}"))


def test_group_remove_user_that_is_not_part_of_group(
    cli_runner: CliRunner,
    developer: User,
    pass_: "PassStore",
    caplog: LogCaptureFixture,
) -> None:
    """
    Given: A group without any user
    When: Removing a user that is not part of the group
    Then: A warning is raised but the program exits fine
    """
    caplog.set_level(logging.INFO)
    pass_.auth.add_user(name=developer.name, email=developer.email, key=developer.key)
    pass_.auth.add_group(name="developers", user_ids=[])

    result = cli_runner.invoke(
        app, ["group", "remove-users", developer.email, "developers"]
    )

    assert result.exit_code == 0
    assert (
        "pass_collaborate.model.auth",
        logging.INFO,
        f"User {developer.name} is not part of the developers group",
    ) in caplog.record_tuples


def test_group_revoke_happy_path(
    cli_runner: CliRunner,
    pass_: "PassStore",
    pass_dev: "PassStore",
    admin: "User",
    developer: "User",
) -> None:
    """
    Given: A configured environment and a group that has access to a password
    When: Revoking access to that directory
    Then: The group no longer has access to the password.
    """
    pass_.auth.add_user(name=developer.name, email=developer.email, key=developer.key)
    pass_.auth.add_group(name="developers", user_ids=[developer.email])
    pass_.change_access("web", ["developers"])
    assert "developers" in pass_.auth.access["web/.gpg-id"]
    assert admin.key in pass_.auth.access["web/.gpg-id"]
    for environment in ("production", "staging"):
        assert pass_.can_decrypt(pass_.path(f"web/{environment}"))
        assert pass_dev.can_decrypt(pass_.path(f"web/{environment}"))

    result = cli_runner.invoke(app, ["group", "revoke", "developers", "web"])

    assert result.exit_code == 0
    pass_.auth.reload()
    assert pass_.auth.access["web/.gpg-id"] == [admin.key]
    for environment in ("production", "staging"):
        assert pass_.can_decrypt(pass_.path(f"web/{environment}"))
        assert not pass_dev.can_decrypt(pass_.path(f"web/{environment}"))


def test_group_revoke_is_idempotent(
    cli_runner: CliRunner, pass_: "PassStore", admin: "User", developer: "User"
) -> None:
    """
    Given: A configured environment
    When: Trying to revoke a directory that the group has no access to
    Then: It doesn't do anything as it's not authorized.
    """
    pass_.auth.add_user(name=developer.name, email=developer.email, key=developer.key)
    pass_.auth.add_group(name="developers", user_ids=[developer.email])
    assert pass_.auth.access == {".gpg-id": [admin.key]}

    result = cli_runner.invoke(app, ["group", "revoke", "developers", "web"])

    assert result.exit_code == 0
    pass_.auth.reload()
    assert pass_.auth.access == {".gpg-id": [admin.key]}



"""Store the classes and fixtures used throughout the tests."""

import shutil
from pathlib import Path

import pytest
from pydantic import EmailStr
from typer.testing import CliRunner

from pass_collaborate.adapters import KeyStore
from pass_collaborate.model.auth import AuthStore, User
from pass_collaborate.model.pass_ import PassStore


@pytest.fixture(name="work_dir")
def work_dir_(tmp_path: Path) -> Path:
    """Create the work directory for the tests."""
    shutil.copytree("tests/assets/pass", tmp_path / ".password-store")
    shutil.copytree("tests/assets/gpg", tmp_path / "gpg")
    return tmp_path


@pytest.fixture(name="pass_")
def pass__(work_dir: Path) -> PassStore:
    """Create the password store for an admin."""
    return PassStore(
        store_dir=(work_dir / ".password-store"), key_dir=(work_dir / "gpg" / "admin")
    )


@pytest.fixture(name="pass_dev")
def pass_dev_(work_dir: Path) -> PassStore:
    """Create the password store for a developer."""
    return PassStore(
        store_dir=(work_dir / ".password-store"),
        key_dir=(work_dir / "gpg" / "developer"),
    )


@pytest.fixture(name="pass_attacker")
def pass_attacker_(work_dir: Path) -> PassStore:
    """Create the password store for an attacker."""
    return PassStore(
        store_dir=(work_dir / ".password-store"),
        key_dir=(work_dir / "gpg" / "attacker"),
    )


@pytest.fixture(name="admin")
def admin_() -> User:
    """Create the admin user."""
    return User(
        name="admin",
        email=EmailStr("admin@example.org"),
        key="5E435D54DDC11A0F303910AEDECE5B3C889F13DE",
    )


@pytest.fixture(name="developer")
def developer_() -> User:
    """Create the developer user."""
    return User(
        name="developer",
        email=EmailStr("developer@example.org"),
        key="8DFE8782CD025ED6220D305115575911602DDD94",
    )


@pytest.fixture(name="attacker")
def attacker_() -> User:
    """Create the attacker user."""
    return User(
        name="Mallory",
        email=EmailStr("mallory@example.org"),
        key="C810FD864F7BAED8AD1D233C6E3A5366E18CBE77",
    )


@pytest.fixture(name="auth")
def auth_(work_dir: Path) -> AuthStore:
    """Create the password store for an admin."""
    auth = AuthStore()
    auth.load(str(work_dir / ".password-store" / ".auth.yaml"))
    return auth


@pytest.fixture(name="key")
def key_(work_dir: Path) -> KeyStore:
    """Create the password store for an admin."""
    return KeyStore(key_dir=work_dir / "gpg" / "admin")


@pytest.fixture(name="cli_runner")
def runner_(work_dir: Path) -> CliRunner:
    """Configure the typer cli runner."""
    return CliRunner(
        env={
            "PASSWORD_STORE_DIR": str(work_dir / ".password-store"),
            "GNUPGHOME": str(work_dir / "gpg" / "admin"),
            "PASSWORD_AUTH_DIR": "",
        },
        mix_stderr=False,
    )



"""Define the data models of the password store."""

import logging
import re
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Generator, List, Optional

from pydantic import BaseModel, root_validator  # noqa: E0611

from ..adapters import KeyStore
from ..exceptions import NotFoundError, TooManyError
from .auth import AuthStore
from .key import GPGKey

if TYPE_CHECKING:
    from .auth import Identifier


log = logging.getLogger(__name__)


class PassStore(BaseModel):
    """Define the adapter of the `pass` password store.

    I've thought of using [passpy](https://github.com/bfrascher/passpy) but it
    doesn't look maintained anymore.

    Args:
        store_dir: Directory where the `pass` password store data lives.
        key_dir: Directory where the `gnupg` data lives.
    """

    store_dir: Path
    key_dir: Path
    auth_dir: Path = Path("")
    # ignore: the keys are going to be set by the root validator
    key: KeyStore = None  # type: ignore
    auth: AuthStore = None  # type: ignore

    class Config:
        """Configure the pydantic model."""

        arbitrary_types_allowed = True

    @root_validator(pre=True)
    @classmethod
    def set_adapters(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Set the adapters."""
        key = KeyStore(key_dir=values["key_dir"])

        auth = AuthStore()
        try:
            if values["auth_dir"]:
                auth_dir = values["store_dir"] / values["auth_dir"]
            else:
                auth_dir = values["store_dir"]
        except KeyError:
            auth_dir = values["store_dir"]
        auth.load(f"{auth_dir}/.auth.yaml")
        cls._load_missing_users_from_keystore(key, auth)
        auth.save()

        values["key"] = key
        values["auth"] = auth
        return values

    @classmethod
    def _load_missing_users_from_keystore(
        cls, keystore: KeyStore, auth: AuthStore
    ) -> None:
        """Load the missing user information from the keystore."""
        missing_keys = []
        for key in cls.all_keys(auth.store_dir):
            try:
                auth.get_user(key)
            except NotFoundError:
                for stored_key in keystore.public_key_fingerprints:
                    if stored_key.match(key):
                        auth.add_user(
                            name=stored_key.name,
                            key=key,
                            email=stored_key.email,
                        )
                        break
                else:
                    missing_keys.append(key)

        if len(missing_keys) > 0:
            raise NotFoundError(
                "Please import the next gpg keys in your gpg keystore:\n"
                + "\n".join(missing_keys)
            )
        auth.save()

    def path(
        self, pass_path: Optional[str] = None, is_dir: Optional[bool] = None
    ) -> Path:
        """Return the path to the file or directory of the password internal path.

        Args:
            pass_path: internal path of the password store. Not a real Path.
            is_dir: if we expect `pass_path` to be a directory.

        Raises:
            ValueError: if is_dir is true and the result path is not a directory.

        Example:
        >>> self.path('web')
        Path('~/.password-store/web')

        >>> self.path()
        Path('~/.password-store')

        >>> self.path('bastion')
        Path('~/.password-store/bastion.gpg')
        """
        if pass_path is None:
            return self.store_dir

        path = Path(self.store_dir / pass_path)

        if is_dir and not path.is_dir():
            raise ValueError(f"{path} is not a directory when it was expected to be")

        if not path.exists():
            path = path.with_name(f"{path.name}.gpg")
            if not path.exists():
                raise NotFoundError(
                    f"Could not find the element {pass_path} in the password store"
                )

        return path

    def allowed_keys(
        self, path: Optional[Path] = None, deep: bool = False
    ) -> List[GPGKey]:
        """Return the allowed gpg keys of the path.

        * For files it analyzes the gpg data if deep is enabled otherwise it
            looks at the parent .gpg-id file like with directories.
        * For directories it traverses the paths to find the first .gpgid file,
            if the auth store has access information of that file it uses that,
            else it will return the content of the gpg-id file.

        If the `path` is None, it will check all keys stored in all .gpg-id
        files in the password store.

        Args:
            path: A real path to an element of the pass store.
            deep: Enable to analyze the keys allowed for each file instead of
                trusting the .gpg-id files.
        """
        # Get existent keys
        if path:
            if path.match("*.gpg") and deep:
                keys = self.key.list_recipients(path)
            else:
                if path.match("*.gpg-id"):
                    gpg_id = path
                else:
                    gpg_id = self.auth.gpg_id_file(path)
                try:
                    keys = self.auth.allowed_keys(str(gpg_id))
                except NotFoundError:
                    keys = gpg_id.read_text().splitlines()
        else:
            keys = []
            for gpg_id in self.store_dir.rglob(".gpg-id"):
                keys.extend(self.allowed_keys(gpg_id))

        return list(set(keys))

    def reencrypt_directory(
        self,
        pass_dir_path: str,
    ) -> None:
        """Reencrypt the password files of a directory to match the desired access.

        Args:
            pass_dir_path: internal path of a directory of the password store.
                Not a real Path

        Raises:
            ValueError: When trying to authorize a file.
                If we authorize a file with keys different than the ones
                specified on the .gpg-id file, the next time someone reencrypts
                the file using `pass` directly, the change will be overwritten.
                We could handle this case, but not for the MVP.

        """
        if self.path(pass_dir_path).is_file():
            raise ValueError(
                "Authorizing access to a file is not yet supported, "
                "please use the parent directory."
            )

        # Re-encrypt all the password files
        for path in self._pass_paths(pass_dir_path):
            log.info(f"Reencrypting file {self._pass_path(path)}")
            # As we're changing the access keys to a file, we can't use
            # self.allowed_keys directly on the file as it will only report the
            # keys that can now access the file.
            allowed_keys = self.allowed_keys(path=self.auth.gpg_id_file(path))
            self.key.reencrypt(path, allowed_keys)

    def find_keys(self, identifier: "Identifier") -> List[GPGKey]:
        """Return the gpg keys associated to an identifier.

        Args:
            identifier: Unique identifier of a group or person. It can be the
                group name, person name, email or gpg key.
        """
        keys = self.auth.find_keys(identifier)
        return keys

    def update_gpg_id_file(
        self,
        gpg_id: str,
    ) -> None:
        """Update the GPG keys of a .gpg-id file to match the auth store access.

        Args:
            gpg_id: path to a password store .gpg-id file
        """
        keys = []
        gpg_id_file = Path(gpg_id)
        log.info(f"Updating the keys stored in {gpg_id}")
        keys.extend(self.allowed_keys(gpg_id_file))
        gpg_id_file.write_text("\n".join(set(keys)), encoding="utf-8")

    def can_decrypt(self, path: Path) -> bool:
        """Test if the user can decrypt a file.

        Args:
            path: Path to the file to decrypt.
        """
        return self.key.can_decrypt(path)

    @property
    def key_id(self) -> str:
        """Return the gpg key id used by the password store user.

        Compare the private keys stored in the keys store with the keys used in
        the password storage.

        Raises:
            NotFoundError: If the user key is not between the allowed keys.
            TooManyError: If the matching algorithm returns more than one key,
                which would be a bug.
        """
        keystore_keys = self.key.private_key_fingerprints
        matching_keys = list(set(keystore_keys) & set(self.allowed_keys()))

        if len(matching_keys) == 1:
            return matching_keys[0]

        if len(matching_keys) == 0:
            log.warning(
                "The user gpg key was not found between the allowed keys in the "
                "password store"
            )
            if len(keystore_keys) == 1:
                return keystore_keys[0]
            raise TooManyError(
                "There is more than one private key in your store and "
                "none is allowed in the password store"
            )
        raise TooManyError(
            "There were more than 1 available gpg keys that is used "
            f"in the repository. Matching keys are: {matching_keys}"
        )

    @staticmethod
    def all_keys(store_dir: Path) -> List[str]:
        """Return all the gpg keys used by the password store."""
        keys = []

        for gpg_id in store_dir.rglob(".gpg-id"):
            keys.extend(gpg_id.read_text().splitlines())

        return list(set(keys))

    def _pass_paths(
        self, pass_dir_path: Optional[str] = None
    ) -> Generator[Path, None, None]:
        """Return all the password files of a pass directory.

        Args:
            pass_dir_path: internal path of a directory of the password store.
                Not a real Path. If None it will take the root of the password
                store
        """
        return self.path(pass_dir_path, is_dir=True).rglob("*.gpg")

    def _pass_path(self, path: Path) -> str:
        """Return the pass representation of a real path.

        It's the reverse of self.path

        Args:
            path: Path to a real directory or file.
        """
        pass_path = re.sub(rf"{self.store_dir}/?", "", str(path))

        if path.is_file():
            pass_path = pass_path.replace(".gpg", "")

        return pass_path

    def change_group_users(
        self,
        group_name: str,
        add_identifiers: Optional[List["Identifier"]] = None,
        remove_identifiers: Optional[List["Identifier"]] = None,
    ) -> None:
        """Change the list of users of an existent group.

        It also reencrypts the passwords associated to that group.

        Args:
            group_name: Group to change
            add_identifiers: Unique identifiers of users to add. It can be user
                names, emails or gpg keys.
            remove_identifiers: Unique identifiers of users to remove. It can
                be the user names, emails or gpg keys.

        Raises:
            EncryptionError: When there was an error reencrypting the files
                when changing the access
        """
        add_identifiers = add_identifiers or []
        remove_identifiers = remove_identifiers or []

        log.debug("Update the group users in the auth store")
        auth_store_changed = self.auth.change_group_users(
            group_name=group_name,
            add_identifiers=add_identifiers,
            remove_identifiers=remove_identifiers,
        )

        if not auth_store_changed:
            return

        log.debug("Reencrypt the passwords that the group has access to")
        start_dir = self.store_dir / self.auth_dir
        for gpg_id in start_dir.rglob(".gpg-id"):
            pass_path = self._pass_path(gpg_id.parent)
            if self.has_access(pass_path, group_name):
                self.change_access(
                    pass_dir_path=pass_path,
                    add_identifiers=add_identifiers,
                    remove_identifiers=remove_identifiers,
                    auth_upstream_change=auth_store_changed,
                )

    def access(self, identifier: "Identifier", deep: bool = False) -> List[str]:
        """Get a list of passwords the entity identified by identifier has access to.

        It only analyzes the passwords that are in the subtree where the
        .auth.yaml file lives.

        Args:
            identifier: Unique identifier of a group or person. It can be the
                group name, person name, email or gpg key.

            deep: Enable to analyze the keys allowed for each file instead of
                trusting the .gpg-id files.

        Returns:
            List of `pass` paths that the entity has access to.

        Raises:
            NotFoundError: If the identifier doesn't match any known GPG keys
        """
        return [
            self._pass_path(path)
            for path in self._pass_paths(str(self.auth_dir))
            if self.has_access(self._pass_path(path), identifier, deep)
        ]

    def has_access(
        self,
        pass_path: str,
        identifier: Optional["Identifier"] = None,
        deep: bool = False,
    ) -> bool:
        """Return if the user of the password store has access to an element of the store.

        If the identifier is None it will assume that we want to check the
        access of the user that initialize the password store.

        * For files it tries to decrypt them.
        * For directories it checks if the active GPG key is in the allowed
            keys of the .gpgid file.

        If identifier is not None, it will check if the entity represented by
        the identified has access to the element in the store.

        * If it's a group it will check if all the gpg keys are allowed and the group is
            allowed in the auth store.
        * If it's a user it will check if it's gpg key is allowed.

        Args:
            pass_path: internal path of the password store. Not a real Path
            identifier: Unique identifier of a group or person. It can be the
                group name, person name, email or gpg key.
            deep: Enable to analyze the keys allowed for each file instead of
                trusting the .gpg-id files.

        Raises:
            NotFoundError: If the identifier doesn't match any known GPG keys
        """
        path = self.path(pass_path)

        log.debug(f"Checking access of {pass_path}")
        # If we want to check the active user
        if not identifier:
            if path.is_file():
                return self.can_decrypt(path)
            keys = [self.key_id]
        else:
            keys = self.auth.find_keys(identifier)

            if len(keys) == 0:
                try:
                    keys = [self.key.find_key(identifier).id_]
                except NotFoundError as error:
                    log.error(
                        f"Coudn't find a valid GPG key for {identifier} in your store"
                    )
                    raise error
        try:
            allowed_keys = self.allowed_keys(path, deep)
            all_keys_allowed = all(key in allowed_keys for key in keys)
        except NotFoundError:
            # if self.key_id raises a NotFoundError is because there is no key that
            # is allowed
            return False

        if not all_keys_allowed:
            return False

        # If it's a group we need to check that the group name is in the access list.
        if identifier:
            try:
                return self.auth.has_access(path, identifier)
            except NotFoundError:
                return False

        return True

    # R0913: too many arguments 6/5, but self shouldn't count :P
    def change_access(  # noqa: R0913
        self,
        pass_dir_path: str,
        add_identifiers: Optional[List["Identifier"]] = None,
        remove_identifiers: Optional[List["Identifier"]] = None,
        ignore_parent: bool = False,
        auth_upstream_change: bool = False,
    ) -> None:
        """Authorize or revoke a group or person to a directory of the password store.

        It will also tweak the `.gpg-id` file to specify the desired access and
        it will reencrypt the files of that directory.

        Args:
            pass_dir_path: internal path of a directory of the password store.
                 Not a real Path
            add_identifiers: Unique identifiers of groups or people to authorize.
                It can be the group name, person name, email or gpg key. Can be
                used in addition to `keys`.
            remove_identifiers: Unique identifiers of groups or people to revoke.
                It can be the group name, person name, email or gpg key. Can be
                used in addition to `keys`.
            ignore_parent: Ignore the access permissions defined in the parent .gpg-id.
                It shouldn't be True by default because it will risk locking
                yourself out.
            auth_upstream_change: If there was a change in the auth done by any
                operations before the call of this function

        Raises:
            ValueError: When trying to authorize a file.
                If we authorize a file with keys different than the ones specified on
                the .gpg-id file, the next time someone reencrypts the file using `pass`
                directly, the change will be overwritten. We could handle this case, but
                not for the MVP.
            EncryptionError: When there was an error reencrypting the files
                when changing the access
        """
        if self.path(pass_dir_path).is_file():
            raise ValueError(
                "Changing access to a file is not yet supported, "
                "please use the parent directory."
            )
        log.info(f"Updating access to {pass_dir_path}")

        gpg_id = f"{self.path(pass_dir_path)}/.gpg-id"

        auth_changed = self.auth.change_access(
            gpg_id=gpg_id,
            add_identifiers=add_identifiers,
            remove_identifiers=remove_identifiers,
            ignore_parent=ignore_parent,
        )

        if auth_changed or auth_upstream_change:
            self.update_gpg_id_file(gpg_id)
            self.reencrypt_directory(pass_dir_path)



"""Define the views of the program."""

from pathlib import Path
from typing import TYPE_CHECKING, Dict, List

from pydantic import BaseModel  # noqa: E0611
from rich.console import Console
from rich.table import Table
from rich.text import Text
from rich.tree import Tree

if TYPE_CHECKING:
    from .model.auth import Group, User


def print_model(model: BaseModel) -> None:
    """Print the attributes of a model.

    Args:
        model: A pydantic model.
    """
    table = Table(box=None, show_header=False)
    table.add_column("Type", justify="center", style="green", no_wrap=True)
    table.add_column("Value")
    for attribute, schema in model.schema()["properties"].items():
        if schema["type"] == "string":
            table.add_row(schema["title"], getattr(model, attribute))
        elif schema["type"] == "array":
            table.add_row(
                schema["title"], "- " + "\n- ".join(getattr(model, attribute))
            )

    Console().print(table)


def print_access(label: str, paths: List[str]) -> None:
    """Print the password paths a user or group has access to.

    Args:
        label: Entity we're printing the access to the password store.
        paths: List of `pass` paths that the entity has access to.
    """
    # Create the tree structure
    tree = Tree(f"Password access for {label}")
    trees: Dict[str, Tree] = {}

    for _path in paths:
        path = Path(_path)
        for parent in list(path.parents)[::-1]:
            if str(parent) == ".":
                active_tree = tree
                continue
            try:
                active_tree = trees[str(parent)]
            except KeyError:
                if str(parent.parent) == ".":
                    active_tree = tree.add(str(parent))
                else:
                    active_tree = active_tree.add(parent.name)
                trees[str(parent)] = active_tree
        active_tree.add(path.name)

    Console().print(tree)


def print_group(group: "Group", users: List["User"]) -> None:
    """Print the information of a group.

    Args:
        group: Group to print
        users: List of Users of the group
    """
    # Create the tree structure
    tree = Tree(group.name)

    for user in users:
        tree.add(
            Text.assemble(
                (user.name, "magenta"),
                ": ",
                (user.email, "green"),
                " (",
                user.key,
                ")",
            ),
        )

    Console().print(tree)



"""Test the command line interface."""

import logging
import os
import re
from typing import TYPE_CHECKING, Dict, List

import pytest
from _pytest.logging import LogCaptureFixture
from typer.testing import CliRunner

from pass_collaborate.entrypoints.cli import app
from pass_collaborate.version import __version__

from ..factories import UserFactory

if TYPE_CHECKING:
    from pass_collaborate.model.auth import User
    from pass_collaborate.model.pass_ import PassStore


log = logging.getLogger(__name__)


def test_version(cli_runner: CliRunner) -> None:
    """Prints program version when called with --version."""
    result = cli_runner.invoke(app, ["--version"])

    assert result.exit_code == 0
    assert re.search(
        rf" *pass_collaborate: {__version__}\n *Python: .*\n *Platform: .*",
        result.stdout,
    )


def test_generates_default_auth_conf_if_none_is_available(
    cli_runner: CliRunner, pass_: "PassStore", admin: "User"
) -> None:
    """
    Given: An environment without an auth config file.
    When: calling an auth related command.
    Then: The config file is created with the template and the users are created from
        The key store.
    """
    os.remove(pass_.auth.config_file)

    result = cli_runner.invoke(app, ["group", "add", "test_group"])

    assert result.exit_code == 0
    assert os.path.exists(pass_.auth.config_file)
    pass_.auth.reload()
    assert pass_.auth.users == [admin]


def test_pass_returns_error_when_no_keys_are_found(
    cli_runner: CliRunner, pass_: "PassStore", caplog: LogCaptureFixture
) -> None:
    """
    Given: A password store that uses a GPG key that is not on our keystore
    When: calling any pass_collaborate command
    Then: An error is returned
    """
    user = UserFactory.build()
    gpg_id = pass_.store_dir / ".gpg-id"
    gpg_id.write_text(user.key)

    result = cli_runner.invoke(app, ["access", user.name])

    assert result.exit_code == 404
    assert (
        f"Please import the next gpg keys in your gpg keystore:\n{user.key}"
        in result.stderr
    )


@pytest.mark.parametrize(
    ("args", "env"),
    [
        (["--auth-dir", "web", "group", "add", "test_group"], {}),
        (["group", "add", "test_group"], {"PASSWORD_AUTH_DIR": "web"}),
    ],
)
def test_supports_auth_file_not_in_root_of_pass(
    cli_runner: CliRunner,
    pass_: "PassStore",
    admin: "User",
    developer: "User",
    args: List[str],
    env: Dict[str, str],
) -> None:
    """
    Given: An environment with an auth config file in a subdirectory of
        `~/.passwordstore`.
    When: calling an auth related command either with the environment variable
        or the flag to set the alternative auth store.
    Then: The config file is created with the template and the users are
        created from The key store.

        The users from the root of the password store are not loaded though,
        only the ones on the children of the specified auth-dir.
    """
    os.remove(pass_.auth.config_file)
    new_auth_store = pass_.auth.store_dir / "web"
    new_config = new_auth_store / ".auth.yaml"
    new_gpg_id = new_auth_store / ".gpg-id"
    new_gpg_id.write_text(developer.key)
    assert not new_config.exists()

    result = cli_runner.invoke(app, args, env=env)

    assert result.exit_code == 0
    assert not os.path.exists(pass_.auth.config_file)
    assert new_config.exists()
    pass_.auth.load(str(new_config))
    assert pass_.auth.users == [developer]
    assert pass_.auth.groups[0].name == "test_group"



"""Define the adapter of the Auth store."""

import logging
from contextlib import suppress
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from goodconf import GoodConf
from pydantic import BaseModel, EmailStr, Field  # noqa: E0611
from ruyaml import YAML

from ..exceptions import NotFoundError, TooManyError
from .key import GPGKey

Name = str
Username = Name
Identifier = Name
GPGIDPath = str


log = logging.getLogger(__name__)


class Group(BaseModel):
    """Model a group of users."""

    name: Name
    users: List[EmailStr] = Field(default_factory=list)

    def match(self, other: Union[str, "Group", "User"]) -> bool:
        """Check if the group or it's users matches other object."""
        if isinstance(other, str):
            return other == self.name or other in self.users
        if isinstance(other, User):
            return other.email in self.users
        return other == self

    def add_users(self, users: List["User"]) -> bool:
        """Add a list of users from the group.

        Returns:
            If there was any user added.
        """
        changed = False

        for user in users:
            if user.email not in self.users:
                log.info(f"Adding user {user.name} to group {self.name}")
                self.users.append(user.email)
                changed = True

        return changed

    def remove_users(self, users: List["User"]) -> bool:
        """Remove a list of users from the group.

        Returns:
            If there was any user removed.
        """
        changed = False
        for user in users:
            log.info(f"Removing user {user.name} from group {self.name}")
            try:
                self.users.remove(user.email)
                changed = True
            except ValueError:
                log.info(f"User {user.name} is not part of the {self.name} group")

        return changed


class User(BaseModel):
    """Model a user of the password store."""

    name: Username
    email: EmailStr
    key: GPGKey

    def match(self, other: Union[str, "Group", "User"]) -> bool:
        """Check if the user matches the identifier."""
        if isinstance(other, str):
            return other in (self.name, self.key, self.email)
        return False


class AuthStore(GoodConf):
    """Define the adapter of the authorisation store."""

    groups: List[Group] = Field(default_factory=list)
    users: List[User] = Field(default_factory=list)
    access: Dict[GPGIDPath, List[Identifier]] = Field(default_factory=dict)

    def add_user(self, name: str, key: str, email: str) -> User:
        """Create a new user.

        Args:
            name: name of the user
            key: GPG key of the user.
            email: Email of the user.

        Raises:
            ValueError: if the user already exists
        """
        for id_, value in [("name", name), ("key", key), ("email", email)]:
            with suppress(NotFoundError):
                user = self.get_user(value)
                if user:
                    raise ValueError(
                        f"The user {user.name} is using the {id_} {value}."
                    )

        new_user = User(name=name, key=key, email=EmailStr(email))
        self.users.append(new_user)
        self.save()
        return new_user

    def add_group(self, name: str, user_ids: Optional[List[str]] = None) -> Group:
        """Create a new group of users.

        Args:
            name: name of the group
            user_ids: user identifiers to add to the group.

        Raises:
            ValueError: if the group already exists
        """
        user_ids = user_ids or []
        user_emails = [self.get_user(id_).email for id_ in user_ids]
        if name in self.group_names:
            raise ValueError(f"The group {name} already exists.")

        new_group = Group(name=name, users=user_emails)
        self.groups.append(new_group)
        self.save()
        return new_group

    @property
    def group_names(self) -> List[str]:
        """Return the names of the groups in the authentication store."""
        return [group.name for group in self.groups]

    @property
    def user_names(self) -> List[str]:
        """Return the names of the users in the authentication store."""
        return [user.name for user in self.users]

    @property
    def config_file(self) -> str:
        """Return the path to the config file."""
        return str(self._config_file)

    @property
    def store_dir(self) -> Path:
        """Return the path to the store directory."""
        return Path(self.config_file).parent

    def reload(self) -> None:
        """Reload the contents of the authentication store."""
        self.load(self.config_file)

    def load(self, filename: Optional[str] = None) -> None:
        """Load a configuration file."""
        if not filename:
            filename = f"{self.store_dir}/.auth.yaml"
        self._check_auth_file(filename)
        super().load(self.config_file)
        self._load_gpg_id_files()

    def _check_auth_file(self, filename: str) -> Path:
        """Return the AuthStore configuration file.

        If the file doesn't exist it will copy the default template.
        """
        config_file = Path(filename).expanduser()
        self._config_file = str(config_file)

        if not config_file.exists():
            self.save()

        return config_file

    def _load_gpg_id_files(self) -> None:
        """Load the data of the gpg-id files that is not already in the access store."""
        for gpg_id in self.store_dir.rglob(".gpg-id"):
            access = self.gpg_id_access_key(gpg_id)
            try:
                existent_keys = self.access_keys(access)
                for key in gpg_id.read_text().splitlines():
                    if key not in existent_keys:
                        self.access[access].append(key)
            except KeyError:
                self.access[access] = gpg_id.read_text().splitlines()

    def save(self) -> None:
        """Save the contents of the authentication store."""
        with open(self.config_file, "w+", encoding="utf-8") as file_cursor:
            yaml = YAML()
            yaml.default_flow_style = False
            yaml.dump(self.dict(), file_cursor)

    def get_group(self, name: str) -> Tuple[Group, List[User]]:
        """Return the group that matches the group name.

        Raises:
            NotFoundError: if no group matches the identifier
        """
        group_match = [group for group in self.groups if group.name == name]

        if len(group_match) == 0:
            raise NotFoundError(f"There is no group that matches {name}.")
        if len(group_match) > 1:
            raise TooManyError(
                f"More than one group matched the selected criteria {name}."
            )

        group = group_match[0]
        users = [self.get_user(id_) for id_ in group.users]

        return group, users

    def get_user(self, identifier: Identifier) -> User:
        """Return the user that matches the user identifier.

        Args:
            identifier: string that identifies the user. It can be either the name,
                the email or the gpg key.

        Raises:
            NotFoundError: if no user matches the identifier
            TooManyError: if more than one user matches the identifier
        """
        user_match = [user for user in self.users if user.match(identifier)]

        if len(user_match) == 0:
            raise NotFoundError(f"There is no user that matches {identifier}.")
        if len(user_match) == 1:
            return user_match[0]
        raise TooManyError(
            f"More than one user matched the selected criteria {identifier}."
        )

    def get_identifier(self, identifier: Identifier) -> Union[Group, User]:
        """Return the group or user that matches the identifier.

        Args:
            identifier: string that identifies the user or group. It can be
                either the name, the email or the gpg key.

        Raises:
            NotFoundError: if no user or group matches the identifier
            TooManyError: if more than one user or group matches the identifier
        """
        # users
        user = None
        with suppress(NotFoundError):
            user = self.get_user(identifier)

        # new groups
        with suppress(NotFoundError):
            group, _ = self.get_group(identifier)
            if user:
                raise TooManyError(
                    f"Both user {user.name} and group {group.name} "
                    f"matched identifier {identifier}, "
                    "cancelling authorization."
                )
            return group
        if user:
            return user
        raise NotFoundError(
            f"No user or group (or more than one) matches identifier {identifier}"
        )

    def find_keys(self, identifier: Identifier) -> List["GPGKey"]:
        """Return the gpg keys that matches the identifier.

        Args:
            identifier: string that identifies a user or group. It can be either the
                name, the email or the gpg key.
        """
        users = []

        # Find in the users
        with suppress(NotFoundError):
            users.append(self.get_user(identifier))

        # Find in the groups
        with suppress(NotFoundError):
            _, group_users = self.get_group(identifier)
            users.extend(group_users)

        return [user.key for user in users]

    def change_group_users(
        self,
        group_name: str,
        add_identifiers: Optional[List[Identifier]] = None,
        remove_identifiers: Optional[List[Identifier]] = None,
    ) -> bool:
        """Change the list of users of an existent group.

        Args:
            group_name: Group to change
            add_identifiers: Unique identifier of a user to add. It can be the
                user name, email or gpg key.
            remove_identifiers: Unique identifier of a user to remove. It can
                be the user name, email or gpg key.

        Returns:
            If there has been applied any change in the auth store
        """
        add_identifiers = add_identifiers or []
        remove_identifiers = remove_identifiers or []
        group, _ = self.get_group(group_name)

        log.debug("Add users")
        users_to_add = [self.get_user(id_) for id_ in add_identifiers]
        added_users = group.add_users(users=users_to_add)

        log.debug("Remove users")
        users_to_remove = [self.get_user(id_) for id_ in remove_identifiers]
        removed_users = group.remove_users(users=users_to_remove)

        log.debug("Remove duplicate information in the access storage")
        for access_entry in self.access.values():
            # If the group we want to change is part of an access entry
            if group.name in access_entry:
                # And if any of the rest of the ids match a user that we have
                # added or removed
                for id_ in access_entry.copy():
                    for user in users_to_add + users_to_remove:
                        if user.match(id_):
                            # Then we can remove that entry from the access
                            # storage as it's already tracked by the group
                            access_entry.remove(id_)
                            break

        self.save()

        if added_users or removed_users:
            return True
        return False

    def change_access(
        self,
        gpg_id: GPGIDPath,
        add_identifiers: Optional[List[Identifier]] = None,
        remove_identifiers: Optional[List[Identifier]] = None,
        ignore_parent: bool = False,
    ) -> bool:
        """Authorize or revoke a group or person to a directory of the password store.

        It will store the access information in the auth store.

        Args:
            gpg_id: path to a passwordstore .gpg-id file
            add_identifiers: Unique identifiers of groups or people to authorize.
                It can be the group name, person name, email or gpg key.
            remove_identifiers: Unique identifiers of groups or people to revoke.
                It can be the group name, person name, email or gpg key.
            ignore_parent: Ignore the access permissions defined in the parent .gpg-id.
                It shouldn't be True by default because it will risk locking
                yourself out.

        Returns:
            Whether a change was made.
        """
        add_identifiers = add_identifiers or []
        remove_identifiers = remove_identifiers or []

        try:
            access = self.access[self.gpg_id_access_key(gpg_id)]
        except KeyError:
            # If the access doesn't exist it will take it's parent as a base
            if not ignore_parent:
                access = self.access[
                    self.gpg_id_access_key(self.gpg_id_file(Path(gpg_id).parent))
                ].copy()
            else:
                access = []

        original_access = access.copy()

        log.debug("Authorize new access")
        for identifier in add_identifiers:
            access = self._authorize_access(identifier, access, gpg_id)

        log.debug("Revoke existing access")
        for identifier in remove_identifiers:
            access = self._revoke_access(identifier, access)

        if access == original_access:
            return False

        self.access[self.gpg_id_access_key(gpg_id)] = access
        self.save()

        return True

    def _authorize_access(
        self, identifier: Identifier, access: List[str], gpg_id: GPGIDPath
    ) -> List[str]:
        """Authorize a new element into an access list.

        Args:
            identifier: Unique identifier of a group or person to authorize.
                It can be the group name, person name, email or gpg key.
            access: the value of an entry of self.access

        Returns:
            updated access with the new element
        """
        new_access = self.get_identifier(identifier)

        if any(self.get_identifier(element).match(new_access) for element in access):
            log.info(
                f"{new_access.name} is already authorized to access "
                f"{gpg_id}, skipping"
            )
            return access
        if isinstance(new_access, User):
            log.info(
                f"  Authorizing access to user {new_access.name}: "
                f"{new_access.email} ({new_access.key})"
            )
        else:
            log.info(f"  Authorizing access to group {new_access.name}")
            log.debug(
                "  Removing the gpg keys that are already tracked by the "
                "group we're adding"
            )
            _, users = self.get_group(identifier)
            for element in access:
                if any(user.match(element) for user in users):
                    log.debug(
                        f"{element} is already included in {identifier} "
                        "removing it from the access entry"
                    )
                    access.remove(element)

        access.append(new_access.name)

        return access

    def _revoke_access(self, identifier: Identifier, access: List[str]) -> List[str]:
        """Revoke an element from an access list.

        Args:
            identifier: Unique identifier of a group or person to revoke.
                It can be the group name, person name, email or gpg key.
            access: the value of an entry of self.access

        Returns:
            updated access without the matching element
        """
        revoke = self.get_identifier(identifier)

        if isinstance(revoke, User):
            log.info(
                f"  Revoking access to user {revoke.name}: "
                f"{revoke.email} ({revoke.key})"
            )
        else:
            log.info(f"  Revoking access to group {revoke.name}")

        # We may pass here when we remove the access of a user to a group,
        # therefore the access to the directory doesn't change as it's
        # binded to the group
        with suppress(ValueError):
            access.remove(revoke.name)

        return access

    def has_access(self, path: Path, identifier: "Identifier") -> bool:
        """Check in the stored access if the identdifier is allowed to the gpg_id file.

        Args:
            path: A real path to an element of the pass store.
            identifier: user or group identifier
        """
        authoree = self.get_identifier(identifier)
        access = self.access[self.gpg_id_access_key(self.gpg_id_file(path))]

        if isinstance(authoree, Group):
            return authoree.name in access

        for id_ in access:
            authorized = self.get_identifier(id_)
            if authorized == authoree:
                return True
            if isinstance(authorized, Group):
                for user_id in authorized.users:
                    if self.get_user(user_id) == authoree:
                        return True
        return False

    def allowed_keys(
        self,
        gpg_id: GPGIDPath,
    ) -> List[GPGKey]:
        """Return the allowed gpg keys of a gpg-id path.

        Args:
            gpg_id: path to a .gpg_id file

        Raises:
            NotFoundError: If there is no data of that gpg-id file.
        """
        try:
            authorees = self.access[self.gpg_id_access_key(gpg_id)]
        except KeyError as error:
            raise NotFoundError(
                f"There is no access information for the gpg-id file {gpg_id}"
            ) from error

        keys = []
        for authoree in authorees:
            if len(authoree) == 40:
                # authoree is a gpg_id key
                keys.append(authoree)
            else:
                keys.extend(self.find_keys(authoree))

        return keys

    def gpg_id_file(self, path: Union[Path, str]) -> Path:
        """Return the first .gpg-id file that applies to a path.

        Args:
            path: A real path to an element of the pass store.
        """
        if isinstance(path, str):
            path = Path(path)

        if path.match("*.gpg-id"):
            return path

        if path.is_file():
            path = path.parent

        gpg_id_path = path / ".gpg-id"

        if gpg_id_path.is_file():
            return gpg_id_path

        if path == self.store_dir:
            raise NotFoundError("Couldn't find the root .gpg-id of your store")

        return self.gpg_id_file(path.parent)

    def gpg_id_access_key(self, gpg_id: Union[Path, str]) -> str:
        """Return the key of the access property for the .gpg-id file.

        It returns the relative path from self.store_dir so that different users
        can use the same .auth.yaml file

        Args:
            gpg_id: A real path to a .gpg_id file.

        Raises:
            ValueError: if the gpg_id path is not part of the store_dir
        """
        if isinstance(gpg_id, str):
            gpg_id = Path(gpg_id)

        return str(gpg_id.relative_to(self.store_dir))

    def access_keys(self, access_id: str) -> List[GPGKey]:
        """Return the GPG keys allowed to an access entry.

        Args:
            access_id: identifier of an access element
        """
        keys: List[GPGKey] = []

        for access in self.access[access_id]:
            authoree = self.get_identifier(access)
            if isinstance(authoree, Group):
                _, users = self.get_group(access)
                keys.extend(user.key for user in users)
            else:
                keys.append(authoree.key)

        return list(set(keys))



"""Define all the orchestration functionality required by the program to work.

Classes and functions that connect the different domain model objects with the adapters
and handlers to achieve the program's purpose.
"""



"""Group command line interface definition."""

from typing import List, Optional

import typer
from rich.console import Console

from .. import exceptions, views

app = typer.Typer()


@app.command()
def add(
    ctx: typer.Context,
    name: str = typer.Argument(
        ...,
        help="name of the group",
    ),
    user_ids: Optional[List[str]] = typer.Argument(
        None,
        help=(
            "Identifiers of the users to add to the group. It can be name, "
            "email or GPG key id."
        ),
    ),
) -> None:
    """Add a new group."""
    auth = ctx.obj["pass"].auth
    auth.add_group(name=name, user_ids=user_ids)


@app.command()
def add_users(
    ctx: typer.Context,
    identifiers: List[str] = typer.Argument(
        ...,
        help=(
            "Unique identifiers of users to add. "
            "It can be user names, emails or gpg keys."
        ),
    ),
    group_name: str = typer.Argument(...),
) -> None:
    """Add a list of users to an existent group."""
    try:
        ctx.obj["pass"].change_group_users(
            group_name=group_name, add_identifiers=identifiers
        )
    except exceptions.EncryptionError as error:
        err_console = Console(stderr=True)
        err_console.print(str(error))
        err_console.print(
            "Please reset the state of your password directory, "
            "fix the origin of the error and try again"
        )
        raise typer.Exit(code=2) from error


@app.command()
def remove_users(  # noqa: B008
    ctx: typer.Context,
    identifiers: List[str] = typer.Argument(
        ...,
        help=(
            "Unique identifiers of users to remove. "
            "It can be user names, emails or gpg keys."
        ),
    ),
    group_name: str = typer.Argument(...),
) -> None:
    """Remove a list of users from an existent group."""
    try:
        ctx.obj["pass"].change_group_users(
            group_name=group_name, remove_identifiers=identifiers
        )
    except exceptions.EncryptionError as error:
        err_console = Console(stderr=True)
        err_console.print(str(error))
        err_console.print(
            "Please reset the state of your password directory, "
            "fix the origin of the error and try again"
        )
        raise typer.Exit(code=2) from error


@app.command(name="list")
def list_(ctx: typer.Context) -> None:
    """List existing groups."""
    auth = ctx.obj["pass"].auth
    print("\n".join(auth.group_names))


@app.command()
def show(ctx: typer.Context, name: str) -> None:
    """Print the information of a group."""
    auth = ctx.obj["pass"].auth
    group, users = auth.get_group(name)
    views.print_group(group, users)


@app.command()
def authorize(  # noqa: B008
    ctx: typer.Context,
    group_name: str,
    pass_paths: List[str] = typer.Argument(
        ..., help="pass directories to give access to."
    ),
    ignore_parent: bool = typer.Option(
        False,
        help="Ignore the access permissions defined in the parent .gpg-id",
    ),
) -> None:
    """Authorize a group to a directory of the password store."""
    pass_ = ctx.obj["pass"]
    err_console = Console(stderr=True)
    for path in pass_paths:
        try:
            pass_.change_access(
                pass_dir_path=path,
                add_identifiers=[group_name],
                ignore_parent=ignore_parent,
            )
        except ValueError as error:
            err_console.print(str(error))
            raise typer.Exit(code=2) from error
        except exceptions.TooManyError as error:
            err_console.print(str(error))
            raise typer.Exit(code=401) from error
        except exceptions.NotFoundError as error:
            err_console.print(str(error))
            raise typer.Exit(code=401) from error


@app.command()
def revoke(  # noqa: B008
    ctx: typer.Context,
    group_name: str,
    pass_paths: List[str] = typer.Argument(
        ..., help="pass directories to give access to."
    ),
) -> None:
    """Revoke a group to a directory of the password store."""
    pass_ = ctx.obj["pass"]
    err_console = Console(stderr=True)
    for path in pass_paths:
        try:
            pass_.change_access(
                pass_dir_path=path,
                remove_identifiers=[group_name],
            )
        except ValueError as error:
            err_console.print(str(error))
            raise typer.Exit(code=2) from error
        except exceptions.TooManyError as error:
            err_console.print(str(error))
            raise typer.Exit(code=401) from error
        except exceptions.NotFoundError as error:
            err_console.print(str(error))
            raise typer.Exit(code=401) from error


if __name__ == "__main__":
    app()



"""User command line interface definition."""

from typing import Optional

import typer
from rich.console import Console

from .. import views
from ..exceptions import NotFoundError

app = typer.Typer()


@app.command()
def add(
    ctx: typer.Context,
    identifier: str = typer.Argument(..., help="GPG identifier"),
    name: Optional[str] = typer.Option(None, help="Overwrite GPG key user name"),
    email: Optional[str] = typer.Option(None, help="Overwrite GPG key email"),
) -> None:
    """Add a new user from the data of the GPG keystore imported keys."""
    pass_ = ctx.obj["pass"]
    try:
        key = pass_.key.find_key(identifier)
    except NotFoundError as error:
        err_console = Console(stderr=True)
        err_console.print(str(error))
        raise typer.Exit(code=404) from error
    if name:
        key.name = name
    if email:
        key.email = email
    pass_.auth.add_user(name=key.name, key=key.id_, email=key.email)


@app.command(name="list")
def list_(ctx: typer.Context) -> None:
    """List existing users."""
    auth = ctx.obj["pass"].auth
    print("\n".join([user.name for user in auth.users]))


@app.command()
def show(
    ctx: typer.Context,
    identifier: str = typer.Argument(
        ..., help="It can be either the name, the email or the gpg key."
    ),
) -> None:
    """Print the information of a user."""
    auth = ctx.obj["pass"].auth
    user = auth.get_user(identifier)
    views.print_model(user)



"""Test the implementation of the access command line command."""

from textwrap import dedent
from typing import TYPE_CHECKING

import pytest
from typer.testing import CliRunner

from pass_collaborate.entrypoints.cli import app

if TYPE_CHECKING:
    from pass_collaborate.model.auth import User
    from pass_collaborate.model.pass_ import PassStore


@pytest.mark.parametrize(
    "identifier",
    [
        "developers",
        "developer@example.org",
        "developer",
        "8DFE8782CD025ED6220D305115575911602DDD94",
    ],
)
def test_access_happy_path(
    cli_runner: CliRunner,
    pass_: "PassStore",
    developer: "User",
    identifier: str,
) -> None:
    """
    Given: A user belonging to a group that has access to some passwords
    When: calling access command line
    Then: A tree is shown with only the elements it has access to
    """
    pass_.auth.add_user(name=developer.name, email=developer.email, key=developer.key)
    pass_.auth.add_group(name="developers", user_ids=[developer.email])
    pass_.change_access(add_identifiers=["developers"], pass_dir_path="web")

    result = cli_runner.invoke(app, ["access", identifier])

    assert result.exit_code == 0
    expected_output = dedent(
        f"""\
        Password access for {identifier}
        └── web
            ├── staging
            └── production
        """
    )
    assert result.stdout == expected_output


def test_access_with_no_groups(
    cli_runner: CliRunner,
    pass_: "PassStore",
    admin: "User",
) -> None:
    """
    Given: An auth store without groups
    When: calling access command line with the admin
    Then: A tree is shown with the elements it has access to
    """
    result = cli_runner.invoke(app, ["access", admin.email])

    assert result.exit_code == 0
    expected_output = dedent(
        f"""\
        Password access for {admin.email}
        ├── bastion
        ├── database
        │   ├── staging
        │   └── production
        └── web
            ├── staging
            └── production
        """
    )
    assert result.stdout == expected_output


@pytest.mark.skip("Not yet")
def test_access_doesnt_analyze_the_files_by_default(
    cli_runner: CliRunner,
    pass_: "PassStore",
    attacker: "User",
    admin: "User",
) -> None:
    """
    Given: An auth store with a file that is encrypted with an attacker's key
        but that's not shown in the .gpg-id file
    When: calling access command line with the attacker id with the deep flag
    Then: The element is not shown as by default it doesn't analyze the files content.

    For the sake of speed.
    """
    pass_.key.encrypt(pass_.store_dir / "bastion.gpg", [admin.key, attacker.key])

    result = cli_runner.invoke(app, ["access", attacker.email])

    assert result.exit_code == 0
    assert "bastion" not in result.stdout


@pytest.mark.skip("Not yet")
def test_access_deep_analyzes_the_files(
    cli_runner: CliRunner,
    pass_: "PassStore",
    attacker: "User",
    admin: "User",
) -> None:
    """
    Given: An auth store with a file that is encrypted with an attacker's key
        but that's not shown in the .gpg-id file
    When: calling access command line with the attacker id with the deep flag
    Then: A tree is shown with the element it has access to
    """
    pass_.key.encrypt(pass_.store_dir / "bastion.gpg", [admin.key, attacker.key])

    result = cli_runner.invoke(app, ["access", "--deep", attacker.email])

    assert result.exit_code == 0
    expected_output = dedent(
        f"""\
        Password access for {attacker.email}
        └── bastion
        """
    )
    assert result.stdout == expected_output



"""Define the fixtures of the e2e tests."""

from pathlib import Path

import pytest
from typer.testing import CliRunner


@pytest.fixture(name="runner")
def fixture_runner(work_dir: Path) -> CliRunner:
    """Configure the Click cli test runner."""
    return CliRunner(
        env={
            "PASSWORD_STORE_DIR": str(work_dir / ".password-store"),
            "GNUPGHOME": str(work_dir / "gpg" / "admin"),
        }
    )



"""Test the KeyStore implementation."""

from pathlib import Path
from typing import TYPE_CHECKING

import pytest
import sh

from pass_collaborate.adapters import KeyStore
from pass_collaborate.exceptions import NotFoundError

if TYPE_CHECKING:
    from pass_collaborate.model.auth import User


def test_create_keystore_raises_exception_if_home_is_not_a_directory() -> None:
    """
    Given: A keystore adapter
    When: initializing it with a directory that doesn't exist
    Then: an exception is thrown
    """
    with pytest.raises(
        NotFoundError,
        match="/tmp/this-doesnt-exist is not a directory that holds gnupg data.",
    ):
        KeyStore(key_dir=Path("/tmp/this-doesnt-exist"))


def test_decrypt_returns_file_not_found(key: KeyStore) -> None:
    """
    Given: A keystore adapter
    When: trying to decrypt a file that doesn't exist.
    Then: raises exception
    """
    with pytest.raises(
        NotFoundError,
        match="Could not find the file to decrypt in /tmp/this-doesnt-exist",
    ):
        key.decrypt(Path("/tmp/this-doesnt-exist"))


def test_can_decrypt_a_file(key: KeyStore, work_dir: Path) -> None:
    """
    Given: A keystore adapter
    When: decrypting a file that we have permission to read
    Then: the output is the expected one
    """
    file_to_decrypt = work_dir / ".password-store" / "bastion.gpg"

    result = key.decrypt(file_to_decrypt)

    assert result == "Qqq*yEbb.]W@c?sDJW&&ym_CR\n"


def test_can_encrypt_a_file(key: KeyStore, work_dir: Path) -> None:
    """
    Given: A keystore adapter
    When: encrypting a file
    Then: the keystore is able to decrypt it and `pass` command too
    """
    text = "this is a test"
    file_to_encrypt = work_dir / ".password-store" / "new_file.gpg"
    file_to_encrypt.write_text(text)

    key.encrypt(file_to_encrypt, [key.private_key_fingerprints[0]])  # act

    assert key.decrypt(file_to_encrypt) == text
    pass_command = sh.Command("pass")
    pass_result = pass_command(
        "show",
        "new_file",
        _env={
            "PASSWORD_STORE_DIR": str(work_dir / ".password-store"),
            "PASSWORD_STORE_GPG_OPTS": f"--homedir {str(work_dir/ 'gpg' / 'admin')}",
        },
    )
    assert pass_result == text


def test_can_reencrypt_a_file(
    key: KeyStore, work_dir: Path, admin: "User", developer: "User"
) -> None:
    """
    Given: A keystore adapter and an encrypted file
    When: reencrypting a file with other keys
    Then: the keystore is able to decrypt it and `pass` command too
    """
    text = "this is a test"
    file_to_encrypt = work_dir / ".password-store" / "new_file.gpg"
    file_to_encrypt.write_text(text)
    key.encrypt(file_to_encrypt, [admin.key])

    key.reencrypt(file_to_encrypt, [admin.key, developer.key])  # act

    assert key.decrypt(file_to_encrypt) == text
    pass_command = sh.Command("pass")
    pass_result = pass_command(
        "show",
        "new_file",
        _env={
            "PASSWORD_STORE_DIR": str(work_dir / ".password-store"),
            "PASSWORD_STORE_GPG_OPTS": f"--homedir {str(work_dir/ 'gpg' / 'admin')}",
        },
    )
    assert pass_result == text


def test_can_list_recipients_of_file(key: KeyStore, work_dir: Path) -> None:
    """
    Given: A keystore adapter
    When: listing the keys that can decrypt a file
    Then: the output is the expected one
    """
    file_to_check = work_dir / ".password-store" / "bastion.gpg"

    result = key.list_recipients(file_to_check)

    assert result == [key.private_key_fingerprints[0]]


def test_find_key_works_with_short_id(key: KeyStore, admin: "User") -> None:
    """
    Given: A keystore adapter
    When: Calling find_key with the short id of a gpg key
    Then: The desired key is found
    """
    result = key.find_key("DECE5B3C889F13DE")

    assert result.id_ == admin.key



"""Test the implementation of the PassStore."""

import shutil
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from pass_collaborate.exceptions import NotFoundError

if TYPE_CHECKING:
    from pass_collaborate.model.auth import User
    from pass_collaborate.model.pass_ import PassStore


@pytest.mark.parametrize(
    ("in_", "out"),
    [
        ("web", "/web"),  # Transform a directory path
        (None, ""),  # Transform None
        ("bastion", "/bastion.gpg"),  # Transform a file path
    ],
)
def test_path_returns_the_expected_results(
    pass_: "PassStore", in_: str, out: str
) -> None:
    """
    Given: A configured PassStore
    When: transforming the incoming paths with the path method
    Then: the expected results are returned
    """
    result = pass_.path(in_)

    assert result == Path(f"{pass_.store_dir}{out}")


def test_path_supports_files_with_dots(pass_: "PassStore") -> None:
    """
    Given: A configured PassStore
    When: transforming a path of a file with dots in the name
    Then: the expected result is returned
    """
    shutil.move(
        str(pass_.store_dir / "bastion.gpg"), str(pass_.store_dir / "the.bastion.gpg")
    )

    result = pass_.path("the.bastion")

    assert result == pass_.store_dir / "the.bastion.gpg"


@pytest.mark.parametrize(
    ("in_", "out"),
    [
        ("web", "web"),  # Transform a directory path
        ("", ""),  # Transform None
        ("bastion.gpg", "bastion"),  # Transform a file path
    ],
)
def test_pass_path_returns_the_expected_results(
    pass_: "PassStore", in_: str, out: str
) -> None:
    """
    Given: A configured PassStore
    When: transforming the incoming paths with the pass_path method
    Then: the expected results are returned
    """
    path = pass_.store_dir / in_
    # W0212 Access to a protected member _pass_path of a client class.

    result = pass_._pass_path(path)  # noqa: W0212

    assert result == out


def test_path_raises_error_if_not_found(pass_: "PassStore") -> None:
    """
    Given: A configured PassStore
    When: transforming the incoming paths with the path method for an inexistent path
    Then: An error is raised
    """
    with pytest.raises(
        NotFoundError,
        match="Could not find the element inexistent_path in the password store",
    ):
        pass_.path("inexistent_path")


def test_authorize_raises_error_if_file(pass_: "PassStore") -> None:
    """
    Given: A configured PassStore
    When: trying to authorize a file instead of a directory
    Then: an error is raised.

    If we authorize a file with keys different than the ones specified on the .gpg-id
    file, the next time someone reencrypts the file using `pass` directly, the change
    will be overwritten. We could handle this case, but not for the MVP
    """
    with pytest.raises(
        ValueError,
        match=(
            "Changing access to a file is not yet supported, "
            "please use the parent directory."
        ),
    ):
        pass_.change_access("bastion", ["developer"])


def test_pass_has_access_to_directory(
    pass_: "PassStore", pass_dev: "PassStore"
) -> None:
    """
    Given: A configured PassStore
    When: checking the access to a directory that only the admin should have access.
    Then: return true for the admin, false for the developer
    """
    result = {
        "admin": pass_.has_access("web"),
        "developer": pass_dev.has_access("web"),
    }

    assert result["admin"]
    assert not result["developer"]


def test_pass_has_access_group_is_not_equal_to_their_keys(
    pass_: "PassStore", admin: "User"
) -> None:
    """
    Given: A configured PassStore and a path whose access is allowed for a gpg key
    When: checking the access to that directory for a group that only has that gpg key
    Then: return false as the user is allowed but not the group
    """
    pass_.auth.add_group(name="developers", user_ids=[admin.email])

    result = pass_.has_access("", "developers")

    assert not result



"""Test the implementation of the AuthStore."""

from typing import TYPE_CHECKING, List

import pytest

from pass_collaborate.exceptions import NotFoundError
from pass_collaborate.model.auth import Group, User

from ...factories import GroupFactory, UserFactory

if TYPE_CHECKING:
    from pass_collaborate.model.auth import AuthStore
    from pass_collaborate.model.key import GPGKey


def test_raises_exception_if_group_exists(auth: "AuthStore") -> None:
    """
    Given: an auth store
    When: creating a group that already exists
    Then: an exception is raised.
    """
    auth.add_group("existent_group")

    with pytest.raises(ValueError, match="The group existent_group already exists."):
        auth.add_group("existent_group")


def test_add_group_without_users(auth: "AuthStore") -> None:
    """
    Given: an auth store
    When: creating a group without users
    Then: the group is created
    """
    result = auth.add_group("No users")

    assert result == Group(name="No users")


def test_get_group_raises_exception(auth: "AuthStore") -> None:
    """
    Given: an auth store
    When: get_group is called on a group that doesn't exist.
    Then: an exception is raised.
    """
    with pytest.raises(
        NotFoundError, match="There is no group that matches fight_club."
    ):
        auth.get_group("fight_club")


def test_raises_exception_if_user_exists(auth: "AuthStore") -> None:
    """
    Given: an auth store
    When: creating a user that already exists
    Then: an exception is raised.
    """
    user = UserFactory.build()
    auth.add_user(user.name, user.key, user.email)

    with pytest.raises(
        ValueError, match=f"The user {user.name} is using the name {user.name}."
    ):
        auth.add_user(user.name, user.key, user.email)


def test_raises_exception_if_a_user_with_same_email_exists(auth: "AuthStore") -> None:
    """
    Given: an auth store
    When: creating a user using an email that is already selected for other user
    Then: an exception is raised.
    """
    user = UserFactory.build()
    user2 = UserFactory.build(email=user.email)
    auth.add_user(user.name, user.key, user.email)

    with pytest.raises(
        ValueError, match=f"The user {user.name} is using the email {user.email}."
    ):
        auth.add_user(user2.name, user2.key, user2.email)


@pytest.mark.parametrize(
    "identifier",
    [
        "developer",
        "developer@example.org",
        "8DFE8782CD025ED6220D305115575911602DDD94",
    ],
)
def test_get_user_happy_path(
    auth: "AuthStore", developer: "User", identifier: str
) -> None:
    """
    Given: A valid user
    When: getting by the email, key or name
    Then: the user is returned
    """
    auth.add_user(developer.name, developer.key, developer.email)

    result = auth.get_user(identifier)

    assert result == developer


@pytest.mark.parametrize(
    ("identifier", "out"),
    [
        ("developer", ["8DFE8782CD025ED6220D305115575911602DDD94"]),
        ("developer@example.org", ["8DFE8782CD025ED6220D305115575911602DDD94"]),
        (
            "8DFE8782CD025ED6220D305115575911602DDD94",
            ["8DFE8782CD025ED6220D305115575911602DDD94"],
        ),
        ("developers", ["8DFE8782CD025ED6220D305115575911602DDD94"]),
        ("not_existent_group_or_user", []),
    ],
)
def test_find_keys_happy_path(
    auth: "AuthStore", developer: "User", identifier: str, out: List["GPGKey"]
) -> None:
    """
    Given: Two valid users and a group
    When: getting by the user email, key or name, or group name, or non matching
    Then: the user is returned
    """
    auth.add_user(developer.name, developer.key, developer.email)
    auth.add_group(name="developers", user_ids=["developer@example.org"])

    result = auth.find_keys(identifier)

    assert result == out


def test_group_remove_users_work_on_non_existent_user() -> None:
    """
    Given: A configured group
    When: Trying to remove a user that is not part of the group
    Then: The command doesn't remove anything but it doesn't raise an exception either
    """
    group = GroupFactory.build()
    user = UserFactory.build()

    result = group.remove_users(users=[user])

    assert result is False


def test_group_add_users_is_idempotent() -> None:
    """
    Given: A configured group
    When: Adding a user that it's already in the store
    Then: The user is not added again
    """
    user = UserFactory.build()
    group = GroupFactory.build(users=[])
    group.add_users([user])

    result = group.add_users([user])

    assert group.users == [user.email]
    assert result is False


def test_user_can_have_accents_on_name() -> None:
    """
    Given: the model of the user
    When: Adding a user with accents in the name
    Then: The user is added well
    """
    user = UserFactory.build(name="Péter")

    result = User(name=user.name, email=user.email, key=user.key)

    assert user == result


def test_auth_loads_gpg_id_even_if_entry_exists(
    auth: "AuthStore", admin: "User"
) -> None:
    """
    Given- : an auth store with a key in the access property
    Then: The missing keys are loaded
    """
    auth.access[".gpg-id"] = []
    auth.save()

    auth.reload()  # act

    assert auth.access[".gpg-id"] == [admin.key]



"""Command line interface definition."""

import logging
import os
import sys
import shutil
from pathlib import Path
from typing import Optional

import typer
from pydantic import ValidationError
from rich.console import Console

from .. import views
from ..exceptions import NotFoundError
from ..model.pass_ import PassStore
from ..version import version_info
from . import group, load_logger, user

log = logging.getLogger(__name__)

app = typer.Typer()
app.add_typer(group.app, name="group")
app.add_typer(user.app, name="user")


def version_callback(value: bool) -> None:
    """Print the version of the program."""
    if value:
        print(version_info())
        raise typer.Exit()


# W0613: version is not used, but it is
# M511: - mutable default arg of type Call, it's how it's defined
# B008: Do not perform function calls in argument defaults. It's how it's defined
# R0913: Too many arguments for the function, but we need them to define the command
#   line interface
@app.callback()
def main(  # noqa: R0913
    ctx: typer.Context,
    version: Optional[bool] = typer.Option(  # noqa: W0613, M511, B008
        None, "--version", callback=version_callback, is_eager=True
    ),
    pass_dir: Path = typer.Option(  # noqa: M511, B008
        "~/.password-store", envvar="PASSWORD_STORE_DIR"
    ),
    key_dir: Path = typer.Option("~/.gnupg", envvar="GNUPGHOME"),  # noqa: M511, B008
    auth_dir: Path = typer.Option(  # noqa: M511, B008
        Path(""),
        envvar="PASSWORD_AUTH_DIR",
        help=(
            "Relative path from the root of the password store to the directory "
            "where the .auth.yaml lives"
        ),
    ),
    verbose: bool = False,
) -> None:
    """A pass extension that helps collectives manage the access to their passwords."""
    ctx.ensure_object(dict)
    err_console = Console(stderr=True)
    load_logger(verbose)
    try:
        ctx.obj["pass"] = PassStore(
            store_dir=pass_dir.expanduser(),
            key_dir=key_dir.expanduser(),
            auth_dir=auth_dir,
        )
    except NotFoundError as error:
        err_console.print(str(error))
        raise typer.Exit(code=404) from error
    except ValidationError as error:
        err_console.print(str(error))
        raise typer.Exit(code=2) from error


@app.command()
def access(
    ctx: typer.Context,
    identifier: str = typer.Argument(
        ...,
        help=(
            "Unique identifier of the user or group who's access to check. "
            "It can be a user name, email, gpg key or group name."
        ),
    ),
    deep: bool = typer.Option(
        False,
        help=(
            "Enable to analyze the keys allowed for each file instead of trusting the "
            ".gpg-id files."
        ),
    ),
) -> None:
    """Check what passwords does the user or group have access to."""
    err_console = Console(stderr=True)
    pass_ = ctx.obj["pass"]
    if deep:
        log.info("If you have many files using --deep may be slow")

    try:
        paths = pass_.access(identifier, deep)
    except NotFoundError as error:
        err_console.print(str(error))
        raise typer.Exit(code=404) from error

    views.print_access(label=identifier, paths=paths)


@app.command()
def init() -> None:
    """Create the needed files in the user's device."""
    if "PASSWORD_STORE_EXTENSIONS_DIR" in os.environ:
        lib = Path(os.environ["PASSWORD_STORE_EXTENSIONS_DIR"])
    else:
        lib = Path.home() / ".password-store" / ".extensions"
    log.debug(f"Creating the {lib} directory")
    os.makedirs(lib, exist_ok=True)

    log.debug("Copying the plugin files")
    package_directory = sys.modules['pass_collaborate'].__path__[0]
    for filename in ["user.bash", "group.bash", "access.bash"]:
        shutil.copyfile(f"{package_directory}/assets/{filename}", lib / filename)
        # nosec: We need the files to be executable
        os.chmod(lib / filename, 0o755)  # nosec

    # Enable the extensions
    for filename in [".bashrc", ".zshrc"]:
        config = Path.home() / filename
        if (
            config.exists()
            and "export PASSWORD_STORE_ENABLE_EXTENSIONS=true" not in config.read_text()
        ):
            log.debug(f"Enabling the pass extensions in {config}")
            with config.open("a") as file_descriptor:
                file_descriptor.write("export PASSWORD_STORE_ENABLE_EXTENSIONS=true")


if __name__ == "__main__":
    app()



"""Define the adapters of the key stores."""

import re
from pathlib import Path
from typing import TYPE_CHECKING, List

from gnupg import GPG
from pydantic import BaseModel

from .exceptions import DecryptionError, EncryptionError, NotFoundError, TooManyError

if TYPE_CHECKING:
    from .model.key import GPGKey


class Key(BaseModel):
    """Model the relevant data of a GPG key."""

    id_: str
    name: str
    email: str
    short_key: str

    def match(self, identifier: str) -> bool:
        """Check if the identifier matches the current key.

        It will check the string against the different key properties.
        """
        return any(
            getattr(self, attribute) == identifier
            for attribute in ["id_", "name", "email", "short_key"]
        )


class KeyStore:
    """Define the adapter of the `gpg` key store."""

    def __init__(self, key_dir: Path, gpg_binary: str = "/usr/bin/gpg2") -> None:
        """Set the gpg connector.

        Args:
            key_dir: Path to the GnuPG home directory where your GPG keys live.

        Raises:
            NotFoundError: If the directory doesn't exist
        """
        key_dir = key_dir.expanduser()
        if not key_dir.is_dir():
            raise NotFoundError(f"{key_dir} is not a directory that holds gnupg data.")
        self.key_dir = key_dir
        self.gpg = GPG(gnupghome=key_dir, gpgbinary=gpg_binary)

    def __repr__(self) -> str:
        """Return a string that represents the object."""
        return f"KeyStore(key_dir={self.key_dir})"

    def decrypt(self, path: Path) -> str:
        """Decrypt the contents of a file.

        Args:
            path: Path to the file to decrypt.

        Raises:
            NotFoundError: if file doesn't exist
            DecryptError: can't decrypt file
        """
        try:
            result = self.gpg.decrypt_file(str(path))
        except ValueError as error:
            raise NotFoundError(
                f"Could not find the file to decrypt in {path}"
            ) from error

        if result.returncode != 0:
            # E1101: Instance of 'Crypt' has no 'stderr' member. But it does
            raise DecryptionError(result.stderr)  # noqa: E1101

        return str(result)

    def can_decrypt(self, path: Path) -> bool:
        """Test if the user can decrypt a file.

        Args:
            path: Path to the file to decrypt.
        """
        try:
            self.decrypt(path)
        except (NotFoundError, DecryptionError):
            return False
        return True

    def encrypt(self, path: Path, keys: List["GPGKey"]) -> None:
        """Encrypt a file for a list of keys.

        Args:
            path: Path to the file to encrypt.
            keys: GPG keys used to encrypt the file.

        Raise:
           EncryptionError: if there is any problem when encrypting the file.
        """
        encrypted_data = self.gpg.encrypt_file(str(path), keys)
        if encrypted_data.ok:
            path.write_bytes(encrypted_data.data)
        else:
            # E1101 Instance of 'Crypt' has no 'stderr' member (no-member). But it does
            raise EncryptionError(encrypted_data.stderr)  # noqa: E1101

    def reencrypt(self, path: Path, keys: List["GPGKey"]) -> None:
        """Reencrypt a file for a list of keys.

        Args:
            path: Path to the file to reencrypt.
            keys: GPG keys used to encrypt the file.

        Raise:
           EncryptionError: if there is any problem when encrypting the file.
        """
        data = self.decrypt(path)
        encrypted_data = self.gpg.encrypt(data, keys)
        if encrypted_data.ok:
            path.write_bytes(encrypted_data.data)
        else:
            # E1101 Instance of 'Crypt' has no 'stderr' member (no-member). But it does
            raise EncryptionError(encrypted_data.stderr)  # noqa: E1101

    def list_recipients(self, path: Path) -> List["GPGKey"]:
        """List the keys that can decrypt a file.

        Args:
            path: Path to the file to check.
        """
        keys = []
        for short_key in self.gpg.get_recipients_file(str(path)):
            try:
                keys.append(self.gpg.list_keys(keys=[short_key])[0]["fingerprint"])
            except IndexError as error:
                raise NotFoundError(
                    f"Could not find gpg key with id {short_key}"
                ) from error

        return keys

    @property
    def private_key_fingerprints(self) -> List[str]:
        """Return the IDs of the private keys."""
        return [key["fingerprint"] for key in self.gpg.list_keys(True)]

    @property
    def public_key_fingerprints(self) -> List[Key]:
        """Return the IDs of the public keys.

        It will only use the first id that it finds.
        """
        keys = []
        for key in self.gpg.list_keys():
            match = re.match(r"(?P<name>.*) <(?P<email>.*)>", key["uids"][0])
            if match is None:
                raise ValueError(
                    "Could not extract the name or email from the gpg key information"
                )
            keys.append(
                Key(
                    id_=key["fingerprint"],
                    name=match["name"],
                    email=match["email"],
                    short_key=key["keyid"],
                )
            )

        return keys

    def find_key(self, identifier: str) -> Key:
        """Return the key that matches the identifier.

        Args:
            identifier: GPG key name, email or fingerprint

        Raises:
            NotFoundError: If the identifier doesn't match any available public keys.
            TooManyError: If more than one available public key match the identifier.
        """
        keys = []
        for key in self.public_key_fingerprints:
            if key.match(identifier):
                keys.append(key)

        if len(keys) == 1:
            return keys[0]
        if len(keys) == 0:
            raise NotFoundError(f"No key found for {identifier}")
        raise TooManyError(f"More than one key matched the identifier {identifier}")



"""Test the implementation of the user command line command."""

import re
from typing import TYPE_CHECKING

import pytest
from typer.testing import CliRunner

from pass_collaborate.entrypoints.cli import app

from ..factories import UserFactory

if TYPE_CHECKING:
    from pass_collaborate.model.auth import AuthStore, User


def test_user_add(cli_runner: CliRunner, auth: "AuthStore", developer: "User") -> None:
    """
    Given: A configured environment
    When: calling user add command
    Then: A user is added
    """
    result = cli_runner.invoke(app, ["user", "add", developer.email])

    assert result.exit_code == 0
    auth.reload()
    saved_user = auth.get_user(developer.key)
    assert developer == saved_user


def test_user_add_can_overwrite_name(
    cli_runner: CliRunner, auth: "AuthStore", developer: "User"
) -> None:
    """
    Given: A configured environment
    When: calling user add command with the --name attribute
    Then: A user is added
    """
    result = cli_runner.invoke(app, ["user", "add", developer.email, "--name", "Marie"])

    assert result.exit_code == 0
    auth.reload()
    saved_user = auth.get_user(developer.key)
    assert saved_user.name == "Marie"


def test_user_add_can_overwrite_email(
    cli_runner: CliRunner, auth: "AuthStore", developer: "User"
) -> None:
    """
    Given: A configured environment
    When: calling user add command with the --email attribute
    Then: A user is added
    """
    result = cli_runner.invoke(
        app, ["user", "add", developer.key, "--email", "other@test.org"]
    )

    assert result.exit_code == 0
    auth.reload()
    saved_user = auth.get_user(developer.key)
    assert saved_user.email == "other@test.org"


def test_user_add_fails_if_gpg_key_not_imported(
    cli_runner: CliRunner, developer: "User"
) -> None:
    """
    Given: a gpg store that doesn't have the gpg key of the user we want to
        import
    When: calling user add command
    Then: An error is returned
    """
    user = UserFactory.build()

    result = cli_runner.invoke(app, ["user", "add", user.key])

    assert result.exit_code == 404
    assert f"No key found for {user.key}" in result.stderr


@pytest.mark.skip("Not yet}")
def test_user_add_works_if_gpg_key_not_imported_but_in_available_keys(
    cli_runner: CliRunner, developer: "User"
) -> None:
    """
    Given: a gpg store that doesn't have the gpg key of the user we want to
        import but the public key is stored in the .gpg-pub-keys directory
    When: calling user add command
    Then: the public key is imported and the user is added well.
    """
    result = False

    assert result


@pytest.mark.skip("Not yet}")
def test_user_add_saves_public_key_in_the_public_key_store(
    cli_runner: CliRunner, developer: "User"
) -> None:
    """
    Given: a gpg store that has the gpg key of the user we want to import but
        the public key storage doesn't
    When: calling user add command
    Then: the public key is added to the storage.
    """
    result = False

    assert result


def test_user_list(cli_runner: CliRunner) -> None:
    """
    Given: A configured environment
    When: calling user list command
    Then: The users are listed
    """
    result = cli_runner.invoke(app, ["user", "list"])

    assert result.exit_code == 0
    assert "admin" in result.stdout


def test_user_show(cli_runner: CliRunner, admin: "User") -> None:
    """
    Given: A configured environment
    When: calling user show command
    Then: The user information is shown
    """
    result = cli_runner.invoke(app, ["user", "show", admin.name])

    assert result.exit_code == 0
    for regexp in [
        rf"Name *{admin.name}",
        rf"Email *{admin.email}",
        rf"Key *{admin.key}",
    ]:
        assert re.search(regexp, result.stdout)
